package asd.td7;

import asd.base.ObjetAClef;

/**
 * Une classe pour l'exercice 4 de la feuille 7.
 */
public class PartitionTernaire {

    public ObjetAClef[] T; // le tableau  partitionner

    /**
     * Construit un objet PartitionTernaire pour le tableau T.
     */
    public PartitionTernaire(ObjetAClef[] T) {
        this.T = T;
    }

    /**
     * Partitionne la tranche T[i..j] et retourne l'indice du pivot.
     * Prcondition : i < j.
     * Valeur rendue : p l'indice du pivot = T[p], i <= p <= j.
     * Postrelation : pour tout x dans T[i..p-1] et tout y dans T[p+1..j]
     * x < T[p] <= y.
     * Complexit : O(j-i).
     */
    private int partitionner(int i, int j) {
	ObjetAClef pivot = T[i];
        int p = i;
        for ( int k = i+1; k <= j; k++ )
            if ( T[k].clef() < pivot.clef() ) {
                p++;
                echanger(p,k);
            }
        // INVARIANT : T[i+1..p] < T[i] et T[p+1..k-1] >= T[i]
        // VARIANT   : k crot et est major par j+1
        echanger(i,p);
        // T[i..p-1] < T[p] <= T[p+1..j]
        return p;
    }

    /**
     * Partitionne la tranche T[i..j] en trois et retourne un tableau P
     * de deux entiers, P[0] = g et P[1] = d tels que :
     * - T[g].clef() <= T[d].clef()
     * - pour tout x, y et z,
     *   x dans T[i..g-1], y dans T[g+1,d-1] et z dans T[d+1..j] on ait : 
     *   x.clef() < T[g].clef() <= T[y].clef() < T[d].clef() <= T[z].clef()
     * Prcondition : i < j
     * Complexit   : THETA(j-i+1).
     */
    public int[] partitionnerEnTrois_1(int i, int j) {
        int[] P = new int[2];
        int k = partitionner(i,j);
        if ( k <= (i+j)/2 ) {
            P[0] = k;
            if ( k+1 < j )
                P[1] = partitionner(k+1,j);
            else
                P[1] = j;
        } else {
            P[1] = k;
            if ( i < k-1 )
                P[0] = partitionner(i,k-1);
            else
                P[0] = i;
        }
        return P;
    }

    /**
     * Partitionne la tranche T[i..j] en trois et retourne un tableau P
     * de deux entiers, P[0] = g et P[1] = d tels que :
     * - T[g].clef() <= T[d].clef()
     * - pour tout x, y et z,
     *   x dans T[i..g-1], y dans T[g+1,d-1] et z dans T[d+1..j] on ait : 
     *   x.clef() < T[g].clef() <= T[y].clef() < T[d].clef() <= T[z].clef()
     * Prcondition : i < j
     * Complexit : THETA(j-i+1).
     */
    public int[] partitionnerEnTrois_2(int i, int j) {
        if ( T[i].clef() > T[j].clef() )
            echanger(i,j);
	int cg = T[i].clef(); int cd = T[j].clef();
        int p = i; int q = j-1; int k = i+1;
        while ( k <= q ) {
            if ( T[k].clef() >= cd ) {
                echanger(k,q); q--;
            }
            else
                if ( T[k].clef() < cg ) {
                    p++; echanger(p,k);
                }
                k++;
        }
        // INVARIANT : T[i+1..p] < T[i] <= T[p+1..k-1] < T[j] <= T[q+1..j]
        // VARIANT   : q - k dcroit et est minor par 0
        echanger(i,p); echanger(k,j);
        // T[i..p-1] < T[p] <= T[p+1..q] < T[k] <= T[k+1..j]
        return new int[]{p,k};
    }
    
    /**
     * Permute T[i] et T[j].
     */
    private void echanger(int i, int j) {
        ObjetAClef tmp = T[i];
        T[i] = T[j];
        T[j] = tmp;
    }
}
